#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include "common.h"

#define LOGBUFSIZE 1024

static int pvr_logfd = -1;

int
pvr_log(char *fmt, ...)
{
	va_list ap;
	char buf[LOGBUFSIZE];
	ssize_t written;
	int len;
	
	if (pvr_logfd < 0)
		return (0);

	va_start(ap, fmt);
	len = vsnprintf(buf, sizeof(buf), fmt, ap);
	va_end(ap);
	if (len < 0)
		return (-1);
	if (len > sizeof(buf) - 1)
		len = sizeof(buf) - 1;
	written = write(pvr_logfd, buf, len);
	if (written != len)
		return (-1);

	return 0;
}

int
pvr_log_init(char *fname)
{
	int fd;

	if ((fd = open(fname, O_WRONLY|O_APPEND|O_CREAT|O_SYNC|O_TRUNC, 0644)) >= 0) {
		pvr_log_deinit();
		pvr_logfd = fd;

		return (0);
	}
	return (-1);
}

int
pvr_log_stdout(void)
{
	if (pvr_logfd < 0)
		return (-1);
	if (dup2(pvr_logfd, STDOUT_FILENO) < 0) {
		LOG("dup2() failed: %s", strerror(errno));
		return (-1);
	}

	return (0);
}

int
pvr_log_deinit(void)
{
	if (pvr_logfd >= 0) {
		close(pvr_logfd);
		pvr_logfd = -1;
	}
	return (0);
}

int
strlcpy(char *dst, const char *src, size_t dst_size)
{
	if (dst_size == 0)
		return (0);
	strncpy(dst, src, dst_size);
	dst[dst_size - 1] = '\0';

	return (0);
}

int
strlcat(char *dst, const char *src, size_t dst_size)
{
	if (dst_size == 0)
		return (0);
	strncat(dst, src, dst_size);
	dst[dst_size - 1] = '\0';

	return (0);
}
