#include <stdio.h>
#include <stdlib.h>
#include <sys/syscall.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <dlfcn.h>
#include <errno.h>
#include <unistd.h>
#include <pthread.h>
#include <inttypes.h>
#include <string.h>
#include <dlfcn.h>
#include "pvr.h"

#define TS_FILE_PATTERN "%s/R%s%03d.ts"
#define MDB_FILE_PATTERN "%s/R%s%03d.mdb"

static char ts_filename[MAX_PATH] = "";
static char mdb_filename[MAX_PATH] = "";
static char recpartname[256];
static char *target;
static int no_dir1 = 0;
static int record_lock;
static int recording = 0;
static int prev_key = -1;

static void *pvr_rec = 0;

void set_io_buffer_size(unsigned val)
{
   static unsigned *pvr_recinstance;

   if (val > 0)
   {
      spIPvr_GetRecInstance(0, &pvr_recinstance);
      if (pvr_recinstance) *(pvr_recinstance + 0x2C/4) = val;
   }
}

static int create_recname()
{
	FILE *testwrite;
	int nr;

	LOG("Creating record file names");

	if (access(config.target1, F_OK) != 0) {
		LOG("Dir[%s] is not accessible", config.target1);
		if (no_dir1 == 0) {
			no_dir1 = 1;
			return -1;
		}
		if (access(config.target2, F_OK) != 0) {
			LOG("Dir[%s] is not accessible", config.target2);
			return -2;
		}
		target = config.target2;
	} else {
		target = config.target1;
		no_dir1 = 0;
	}
	LOG("Selected dir[%s]", target);

	EPG_get_data(recpartname, sizeof(recpartname));
	LOG("Got epg data[%s]", recpartname);

	for (nr = 1; nr < 999; nr++) {
		snprintf(ts_filename, sizeof(ts_filename),
				 "%s/R%s%03d.ts", target, recpartname, nr);
		if (access(ts_filename, F_OK) != 0) {
			if (config.testwrite) {
				testwrite = fopen(ts_filename, "a+");
				if (testwrite) {
					fclose(testwrite);
					remove(ts_filename);
				} else
					continue;
			}
			sprintf(mdb_filename, "%s/R%s%03d.mdb", target, recpartname, nr);
			return nr;
		}
	}
	return (target == config.target1)?-1:-2;
}

#if 0
/* pvr_rec_set_test is an only function to debug spI_Stream_Record() */
int pvr_rec_set_test(void *pvr_rec, struct pvr_rec_set_info *info, int flag) __attribute__ ((noinline));

int pvr_rec_set_test(void *pvr_rec, struct pvr_rec_set_info *info, int flag)
{
 	LOG("fname[%s] xC8[%X] xC4[%X] xC0[%X] xBC[%X] xB8[%X] u10[%X] u11[%X] u12[%X] "
		"ap[%X] u20[%X] u21[%X] x9C[%X] vp[%X] u30[%X] u31[%X] u32[%X] u33[%X] p[%X] u4[%X] "
		"x7C[%X] x70[%X] u50[%X] u51[%X] x64[%X] u6[%X] x5C[%X] "
		"u70[%X] u71[%X] u72[%X] u73[%X] u74[%X] u75[%X] u76[%X] u77[%X] u78[%X] u79[%X] "
		"x30[%X] u80[%X] u81[%X] u82[%X] u83[%X]",
 		info->fname, info->xC8, info->xC4, info->xC0, info->xBC, info->xB8,
		info->unk1[0], info->unk1[1], info->unk1[2], info->audio_pid, info->unk2[0],
		info->unk2[1], info->x9C, info->video_pid,
		info->unk3[0], info->unk3[1], info->unk3[2], info->unk3[3], info->pcr_pid, info->unk4,
		info->x7C, info->x70, info->unk5[0], info->unk5[1], info->x64, info->unk6, info->x5C,
		info->unk7[0], info->unk7[1], info->unk7[2], info->unk7[3], info->unk7[4],
		info->unk7[5], info->unk7[6], info->unk7[7], info->unk7[8], info->unk7[9],
		info->x30, info->unk8[0], info->unk8[1], info->unk8[2], info->unk8[3]);

	return (-1);
}
#endif

/* Bugged spI_Stream_Record() replacement */
static int
pvr_stream_record(char *filename)
{
	struct pvr_rec_set_info record_info;
	int fname_len;
	int fd = -1;
	int rc = -1;

	fname_len = strlen(filename);
	if (fname_len >= sizeof(record_info.fname)) {
		LOG("Too long filename[%s]: %d >= %d",
			filename, fname_len,
			(int)sizeof(record_info.fname));
		goto DONE;
	}

	bzero(&record_info, sizeof(record_info));
	strlcpy(record_info.fname, filename, sizeof(record_info.fname));

	record_info.video_pid = 1;
	record_info.audio_pid = 2;
	record_info.pcr_pid = 3;

	/*
     * Fill record_info structure with magic values got from
     * pvr_rec_set_test(). I think most of the big numbers are just
     * uninitialized variables of spI_Stream_Record() but they
     * shouldn't be initialized with zero or spIPvr_RecSet() will
     * fail.
     */
	record_info.xC8 = 1;
	record_info.xC4 = 0;
	record_info.xC0 = 0;
	record_info.xBC = 0;
	record_info.x5C = 0;
/* 	record_info.x30 = 1; */
	record_info.x30 = 0x71E113EC;
	record_info.xB8 = 1;

	record_info.unk2[0] = 0;
	record_info.unk2[1] = 0x1000;
	record_info.unk3[0] = 0x40884000;
	record_info.unk3[1] = 0x2;
	record_info.unk3[2] = 0x407CF8B8;
	record_info.unk5[0] = 0xFFFFFFFF;
	record_info.unk5[1] = 0;
	record_info.x9C = 0;
/* 	record_info.x64 = 0; */
 	record_info.x64 = 1;
	record_info.unk6 = 0x2D64550;
	record_info.x7C = 0;
/* 	record_info.x70 = 0; */
	record_info.x70 = 0x2D64550;
	record_info.unk7[0] = 1;
	record_info.unk7[2] = 0x1008;
	record_info.unk7[3] = 0x40885054;
	record_info.unk7[4] = 0x2D645F0;
	record_info.unk7[5] = 0x40884000;
	record_info.unk7[6] = 0x40883D10;
	record_info.unk7[9] = 0x7170E5DC;
	record_info.unk8[0] = 0x1A;
	record_info.unk8[1] = 0x71E11B01;
	record_info.unk8[2] = 0x71E11C2C;
	record_info.unk8[3] = 0x49;

	if ((fd = open(record_info.fname, O_RDWR | O_CREAT, 0666)) < 0) {
		LOG("Cannot open file[%s]: %s", record_info.fname, strerror(errno));
		goto DONE;
	}
	if ((rc = spIPvr_RecOpen(0, &pvr_rec)) != 0) {
		LOG("Cannot open PVR rec device: %x", rc);
		goto DONE;
	}
	LOG("Opened PVR rec device[%p]", pvr_rec);
	if ((rc = spIPvr_RecSet(pvr_rec, &record_info, 1)) != 0) {
		LOG("Cannot set PVR rec device[%p]: %X", pvr_rec, rc);
		goto DONE;
	}
	LOG("Set PVR rec device[%p]", pvr_rec);
	if ((rc = spIPvr_RecStart(pvr_rec)) != 0) {
		LOG("Cannot start PVR rec device[%p]: %X", pvr_rec, rc);
		goto DONE;
	}
  DONE:
	if (rc == 0)
		LOG("Started recording to[%s] using PVR rec device[%p]",
			record_info.fname, pvr_rec);
	if (fd != -1)
		close(fd);
	return (rc == 0) ? 0 : -1;
	
}

/* spI_Stream_Record_Stop() replacement */
static int
pvr_stream_record_stop(void)
{
	int rc;

	if (pvr_rec == NULL) {
		LOG("Not in record state!");
		return (-1);
	}
	if ((rc = spIPvr_Rec_UnregisterCallback(pvr_rec, 1)) != 0) {
		LOG("Failed to unregister record stream callback "
			"on device[%"PRIxPTR"]: %x", (uintptr_t)pvr_rec, rc);
		return (-1);
	}
	if ((rc = spIPvr_RecStop(pvr_rec)) != 0) {
		LOG("Failed to stop PVR record device[%"PRIxPTR"]: %x",
			(uintptr_t)pvr_rec, rc);
		return (-1);
	}
	if ((rc = spIPvr_RecClose(pvr_rec)) != 0) {
		LOG("Failed to close PVR record device[%"PRIxPTR"]: %x",
			(uintptr_t)pvr_rec, rc);
		return (-1);
	}
	LOG("Stopped recording with PVR rec device[%"PRIxPTR"]",
		(uintptr_t)pvr_rec);
	pvr_rec = NULL;

	return (0);
}

int button_pressed(int key) __attribute__ ((noinline));
int
button_pressed(int key)
{
	struct stat st;
	int deleted;
	int rc;

	int key_really_pressed = key;

	if (config.record_key_lock > 0) {
		if (key == KEY_STOP && prev_key == KEY_STOP) {

			key = KEY_NOTHING;
			record_lock++;
			if (record_lock == 4) {
				key = show_WARNING(ICON_LOCKED);
				LOG("Locked shown, ret[0x%X]", key);
			} else if (record_lock == 8) {
				if (config.record_key_lock == 3)
					record_lock--;
				else {
					key = show_WARNING(ICON_UNLOCKED);
					record_lock = 0;
					LOG("Unlocked shown, ret[0x%X]", key);
				}
			}
			LOG("Stop key, record lock now is[%d]", record_lock);
		} else if (record_lock < 4) {
			record_lock = 0;
			LOG("Set record lock to zero");
		} else if (record_lock < 8) {
			record_lock = 4;
			LOG("Set record lock to 4");
		}
	}

	switch (key) {
	case KEY_STOP:
	case KEY_POWER:
		LOG("Stop key pressed[0x%X]", key);
		if (!recording) {
			LOG("We're not recording");
			break;
		}
		recording = 0;


#if 0
		spI_Stream_Record_Stop();
		rc = 0;
#endif
		rc = pvr_stream_record_stop();
		LOG("recording stopped, rc[%d]", rc);

		if (mdb_filename[0]) {
			if (config.del_mdb) {
				remove(mdb_filename);
				LOG("Removed mdb file[%s]", mdb_filename);
			}
			mdb_filename[0]=0;
		}

		deleted = 0;
		if (ts_filename[0] && stat(ts_filename, &st) == 0) {
			if (config.del_empty_ts && st.st_size == 0) {
				remove(ts_filename);
				deleted = 1;
				LOG("Removed empty ts file[%s]", ts_filename);
			} else
				LOG("Written %"PRIu64" bytes to file[%s]",
					(uint64_t)st.st_size, ts_filename);
		}

		if (key == KEY_STOP)
			key = hide_REC(deleted);
		no_dir1 = 0;
		break;

	case KEY_RECORD:
		LOG("Record key pressed[0x%X]", key);
		if (recording) {
			key = show_FORBIDDEN(target == config.target1 ? 1 : 2);
			LOG("We're already recording, dir[%s], ret[0x%X]", target, key);
		} else if ((config.record_key_lock > 0) && (record_lock >= 4)) {
			key = show_WARNING(ICON_LOCKED);
			LOG("Recording is locked, ret0x%X]", key);
		} else {
			int cn = create_recname();

			if (cn > 0) {

#if 0
				spI_Stream_Record(ts_filename);
#endif
				if ((rc = pvr_stream_record(ts_filename)) != 0) {
					LOG("Cannot record to[%s]", ts_filename);
					key = show_WARNING((target == config.target1)?ICON_NOREC1:ICON_NOREC2);
				} else {
					recording = 1;

					LOG("Started recording to[%s], rc[%d] ret[0x%X]",
						ts_filename, rc, key);
					key = show_REC(target == config.target1 ? 1 : 2);
				}
			}
			else {
				key = show_WARNING((cn==-1)?ICON_NOREC1:ICON_NOREC2);
				LOG("Cannot recording, cn[%d] ret[0x%X]", cn, key);
			}
		}
		break;

	case KEY_PAUSE:
		LOG("Pause key pressed[0x%X]", key);
		if (recording) key = icon_displayed()?hide_REC(0):show_REC(target == config.target1 ? 1 : 2);
		break;

	case KEY_SOURCE:
		if (recording) {
			LOG("Forbidden source key pressed[0x%X]", key);
			key = show_FORBIDDEN(target == config.target1 ? 1 : 2);
		} else if (config.default_source != 0 &&
				   prev_key != config.default_source && prev_key != KEY_SOURCE)
			key = config.default_source;

		break;

	case KEY_FAV_CH:
		LOG("Fav.ch key pressed[0x%X]", key);

		if (prev_key != KEY_STOP) {
			if (config.aspect_ratio)
				key = KEY_ASPECT;
			else if (recording)
				key = show_FORBIDDEN(target == config.target1 ? 1 : 2);
		} else if (recording)
			key = show_FORBIDDEN(target == config.target1 ? 1 : 2);
		break;

	case KEY_P_UP:
	case KEY_P_DOWN:
	case KEY_PRE_CH:
	case KEY_CH_LIST:
	case KEY_TV:
	case KEY_NUM0:
	case KEY_NUM1: case KEY_NUM2: case KEY_NUM3:
	case KEY_NUM4: case KEY_NUM5: case KEY_NUM6:
	case KEY_NUM7: case KEY_NUM8: case KEY_NUM9:
	case KEY_CONTENT:
	case KEY_MEDIA_P:
	case KEY_TTX_MIX:
		LOG("Forbidden key pressed[0x%X]", key);

		if (recording && config.protect) key = show_FORBIDDEN(target == config.target1 ? 1 : 2);
		break;

	default:
		break;
	}

  prev_key = key_really_pressed;
  return(key);
}

void injection(int a, int key, int arg_r2) __attribute__ ((noinline));
void injection(int a, int key, int arg_r2)
{
	// gcc should put "LDMFD SP!, {R4,LR}" on top, if everything goes as planned
	asm volatile(
		"STMFD   SP!, {R0,R2,R3,R12}\n"
	);

	button_pressed(key);
	
	asm volatile(
		"MOV     R1, R0\n" // remapped keycode
		"MOV     R6, R0\n" // remapped keycode
		"LDMFD   SP!, {R0,R2,R3,R12}\n"
		"LDMFD   SP!, {R4,LR}\n"
		"MOV     R7, R2\n"  //This is original command that we overwrite
		"LDR     R3, [R8]\n"  //This is original command that we overwrite
		"LDR     PC, =(_ZN9KeyCommon17SendKeyPressInputEii + 0x18 )\n"	//return after injection
	);
}

void pvr_config(const char *path)  __attribute__ ((noinline));
void pvr_config(const char *path)
{
	load_config(path);

	if (!config.debug) {
		LOG("Logging is disabled");
		pvr_log_deinit();
	}
	LOG("PVR initialized");

	if (config.debug && config.exedsp_debug) {
		pvr_log_stdout();
		/* With printlevel=0xffffffff the TV will hang: too many logging */
		spI_SetPrintLevel(0x100000);
		spI_SetPrintModule(0xffffffff);
	}

    set_io_buffer_size(config.io_buffer);
    if (config.record_key_lock == 2)
		record_lock = 4;
    else
		record_lock = 0;
}
